// Heatbugs application. Copyright (C) 1996-1999 Santa Fe Institute.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

#import "ObserverSwarm.h"
#import "ModelSwarm.h"
#import <collections.h>
#import <objectbase.h>
#import <analysis.h>
#import <gui.h>
#import "ListWrapper.h"
#import "WrappedListGrid2d.h"

@implementation ObserverSwarm

// createBegin: here we set up the default observation parameters.

+ createBegin: aZone
{
  ObserverSwarm *obj;
  id <ProbeMap> probeMap;
  
  // Superclass createBegin to allocate ourselves.

  obj = [super createBegin: aZone];

  // Fill in the relevant parameters (only one, in this case).

  obj->displayFrequency = 1;

  // Also, build a customized probe map. Without a probe map, the default
  // is to show all variables and messages. Here we choose to
  // customize the appearance of the probe, give a nicer interface.

  probeMap = [EmptyProbeMap createBegin: aZone];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  // Add in a bunch of variables, one per simulation parameters

  [probeMap addProbe: [probeLibrary getProbeForVariable: "displayFrequency"
				    inClass: [self class]]];

//   [probeMap addProbe: [[probeLibrary getProbeForMessage: "graphBug:"
// 			     inClass: [self class]]
// 			setHideResult: YES]];

  // Now install our custom probeMap into the probeLibrary.

  [probeLibrary setProbeMap: probeMap For: [self class]];

  return obj;
}

// createEnd: create objects we know we'll need. In this case, none,
// but you might want to override this.

- createEnd
{
  return [super createEnd];
}

- _worldRasterDeath_ : caller
{
  [worldRaster drop];
  worldRaster = nil;
  return self;
}

- _unhappyGraphDeath_ : caller
{
  [unhappyGraph drop];
  unhappyGraph = nil;
  return self;
}

// Create the objects used in the display of the model. This code is
// fairly complicated because we build a fair number of widgets. It's
// also a good example of how to use the display code.

- buildObjects
{
  int i;

  [super buildObjects];
  
   modelSwarm = [ModelSwarm create: self];

  CREATE_ARCHIVED_PROBE_DISPLAY (modelSwarm);
  CREATE_ARCHIVED_PROBE_DISPLAY (self);
 

  [controlPanel setStateStopped];


  [modelSwarm buildObjects];

  // Now get down to building our own display objects.

  // First, create a colormap: this is a global resource, the information
  // here is used by lots of different objects.

  colormap = [Colormap create: self];

  // Colours [0,64) are assigned to the range Red [0, 1), for heat display.

  for (i = 0; i < 64; i++)
    [colormap setColor: i ToRed: (double)i / 63.0 Green: 0 Blue: 0];

  // Colour 64 is set to green, to display heatbugs

  [colormap setColor: 64 ToName: "green"];

  // Colour 65 is set to white, used in this case below on probed heatbug.

  [colormap setColor: 65 ToName: "white"];

  for(i=66; i < 131; i++)
    [colormap setColor: i  ToRed: 0  Green: (double)(i-66) / 65.0 Blue: ((double)(i-66)/65)+0.5  ];


  // Now go in to the heatbugs in the model and set their colours to green (64)

  [[modelSwarm getCitizenList] forEach: M(setColor:) : (void *) 64];
  
  // Next, create a 2d window for display, set its size, zoom factor, title.
  
  worldRaster = [ZoomRaster createBegin: self];
  SET_WINDOW_GEOMETRY_RECORD_NAME (worldRaster);
  worldRaster = [worldRaster createEnd];
  [worldRaster enableDestroyNotification: self
               notificationMethod: @selector (_worldRasterDeath_:)];
  [worldRaster setColormap: colormap];
  [worldRaster setZoomFactor: 4];
  [worldRaster setWidth: [[modelSwarm getPositionGrid] getSizeX]
	       Height: [[modelSwarm getPositionGrid] getSizeY]];
  [worldRaster setWindowTitle: "Protest Grid"];
  [worldRaster pack];				  // draw the window.

  // Now create a Value2dDisplay: this is a special object that will
  // display arbitrary 2d value arrays on a given Raster widget.

  discontentDisplay = [Value2dDisplay createBegin: self];
  [discontentDisplay setDisplayWidget: worldRaster colormap: colormap];
  [discontentDisplay setDiscrete2dToDisplay: [modelSwarm getDiscontentGrid]];
  [discontentDisplay setDisplayMappingM: 2 C: 0];  // turn [0,32768) -> [0,64)
  discontentDisplay = [discontentDisplay createEnd];
 
  protestDisplay = [Object2dDisplay createBegin: self];
  [protestDisplay setDisplayWidget: worldRaster];
  [protestDisplay setDiscrete2dToDisplay: [modelSwarm getPositionGrid]];
  //[protestDisplay setObjectCollection: [modelSwarm getCitizenList]];
  [protestDisplay setDisplayMessage: M(drawSelfOn:)];   // draw method
  protestDisplay = [protestDisplay createEnd];

  // Also, tell the world raster to send mouse clicks to the heatbugDisplay
  // this allows the user to right-click on the display to probe the bugs.

  [worldRaster setButton: ButtonRight
	     /*  Client: protestDisplay*/ Client: [modelSwarm getPositionGrid]
               Message: M(makeProbeAtX:Y:)];

  // Create the graph widget to display unhappiness.

  unhappyGraph = [EZGraph createBegin: self];
  SET_WINDOW_GEOMETRY_RECORD_NAME (unhappyGraph);
  [unhappyGraph setTitle: "Unhappiness of Citizens vs. time"];
  [unhappyGraph setAxisLabelsX: "time" Y: "unhappiness"];
  unhappyGraph = [unhappyGraph createEnd];
  [unhappyGraph enableDestroyNotification: self
                notificationMethod: @selector (_unhappyGraphDeath_:)];

  [unhappyGraph createAverageSequence: "Disgust With Rulers"
                         withFeedFrom: [modelSwarm getCitizenList] 
                          andSelector: M(getUnhappiness)];
  [unhappyGraph createSequence: "Number Of Protesters"
		withFeedFrom:  [modelSwarm getDiscontentGrid]
                          andSelector: M(getPctProtesting)];

  return self;
}  

- _update_
{
  if (worldRaster)
    {
      
      [discontentDisplay display];
      [protestDisplay display];
      [worldRaster drawSelf];
#ifdef SCREENSHOTS
      {
        char filename[40];
        if(getCurrentTime() % 20==0)
	  {
        sprintf (filename, "graph%07d.png", getCurrentTime ());
        [actionCache doTkEvents];
        [[[[[[Pixmap createBegin: self]
              setWidget: [unhappyGraph getGraph]]
             setDecorationsFlag: NO]
            createEnd] save: filename] drop];
        sprintf (filename, "raster%07d.png", getCurrentTime ());
        [[[[[[Pixmap createBegin: self]
              setWidget: worldRaster]
             setDecorationsFlag: YES]
            createEnd] save: filename] drop];
	  }
	  }
#endif
    }
  if (unhappyGraph)
    [unhappyGraph step];
  return self;
}

// Create the actions necessary for the simulation. This is where
// the schedule is built (but not run!)
// Here we create a display schedule - this is used to display the
// state of the world and check for user input. This schedule should
// be thought of as independent from the model - in particular, you
// will also want to run the model without any display.

- buildActions
{
  [super buildActions];
  
  // First, let our model swarm build its own schedule.

  [modelSwarm buildActions];
  
  // Create an ActionGroup for display: a bunch of things that occur in
  // a specific order, but at one step of simulation time. Some of these
  // actions could be executed in parallel, but we don't explicitly
  // notate that here.
  
  displayActions = [ActionGroup create: self];

  // Schedule up the methods to draw the display of the world
  // Also schedule the update of the unhappiness graph
  //  [displayActions createActionTo: unhappinessAverager message: M(update)];

  [displayActions createActionTo: self message: M(_update_)];

  // Schedule the update of the probe displays

  [displayActions createActionTo: probeDisplayManager message: M(update)];

  // Finally, schedule an update for the whole user interface code.
  // This is crucial: without this, no graphics update and the control
  // panel will be dead. It's best to put it at the end of the display schedule

  [displayActions createActionTo: actionCache message: M(doTkEvents)];

  // And the display schedule. Note the repeat interval is set from our
  // own Swarm data structure. Display is frequently the slowest part of a
  // simulation, so redrawing less frequently can be a help.

  displaySchedule = [Schedule createBegin: self];
  [displaySchedule setRepeatInterval: displayFrequency]; // note frequency!
  displaySchedule = [displaySchedule createEnd];
  [displaySchedule at: 0 createAction: displayActions];

  stopSchedule = [Schedule create: [self getZone]];
  [stopSchedule at: 150
    createActionTo: self
         message: M(stopRunning)];


  
  return self;
}  

// activateIn: - activate the schedules so they're ready to run.
// The swarmContext argument has to do with what we were activated *in*.
// Typically the ObserverSwarm is the top-level Swarm, so it's activated
// in "nil". But other Swarms and Schedules and such will be activated
// inside of us.

- activateIn:  swarmContext
{
  // First, activate ourselves (just pass along the context).

  [super activateIn: swarmContext];

  // Activate the model swarm in ourselves. The model swarm is a
  // subswarm of the observer swarm.

  [modelSwarm activateIn: self];

  // Now activate our schedule in ourselves. This arranges for the
  // execution of the schedule we built.

  [displaySchedule activateIn: self];
 
  [stopSchedule activateIn: self]; 
  // Activate returns the swarm activity - the thing that's ready to run.

  return [self getActivity];
}


-stopRunning

{
  //    [getTopLevelActivity() terminate];
      [[modelSwarm getActivity] terminate];
  
      [[self getActivity] terminate];
      [controlPanel setStateQuit]; 
	return self;
}



// You could override the "go" method here if you want something special
// to happen when the model and observer actually start running. But
// the default GUISwarm go is probably good enough.

// - graphBug: aBug
// {
//   if (unhappyGraph)
//     [unhappyGraph createSequence: "Bug" 
//                   withFeedFrom: aBug 
//                   andSelector: M(getUnhappiness)];
//   return self;
// }

@end
